#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <errno.h>

#include "cross_utils.h"

#include <fstream>
#include <iostream>
#include "NvJpegDecoder.h"
#include "NvJpegEncoder.h"

#define ANCIL_FD_BUFFER(n) \
    struct { \
	struct cmsghdr h; \
	int fd[n]; \
    }


CrossUtils::CrossUtils() :
    m_nSocketfd(-1),
    m_nCount(0)
{
    int i;    
    for(i=0;i<MAX_CLIENT;i++)  
    {
        m_nClientfd[i]=0;
        m_nFlag[i]=0;
    }
}

CrossUtils::~CrossUtils()
{
}

uint32_t CrossUtils::get_time_diff_ms(const struct timeval *start, const struct timeval *end)
{
    return (uint32_t)end->tv_sec*1000+(uint32_t)end->tv_usec/1000 -((uint32_t)start->tv_sec*1000+(uint32_t)start->tv_usec/1000);
}

int CrossUtils::ServerCreate()
{
    struct sockaddr_un sock_addr;
	
    m_nSocketfd = socket(PF_UNIX, SOCK_STREAM, 0);
    if (m_nSocketfd < 0) {
        printf("%s: socket create failed\n", __func__);
        return -1;
    }

    unlink(SOCK_PATH);
    memset(&sock_addr, 0, sizeof(struct sockaddr_un));
    sock_addr.sun_family = AF_UNIX;
    strncpy(sock_addr.sun_path,
            SOCK_PATH,
            sizeof(sock_addr.sun_path)-1);

    if (bind(m_nSocketfd,
             (const struct sockaddr*)&sock_addr,
             sizeof(struct sockaddr_un))) {
		perror("bind");		 
        //printf("%s: bind error\n", __func__);
		close(m_nSocketfd);
        return -1;
    }

    if (listen(m_nSocketfd,MAX_CLIENT)) {
		perror("listen");	
        //printf("%s: listen error\n", __func__);
		close(m_nSocketfd);
        return -1;
    }
    printf("%s: m_nSocketfd:%d\n",__func__,m_nSocketfd);

    return m_nSocketfd;
}


int CrossUtils::ClientCreate()
{
    struct sockaddr_un sock_addr;
    int wait_loop = 0;
    
    m_nSocketfd = socket(PF_UNIX, SOCK_STREAM, 0);
    if(m_nSocketfd < 0) {
        printf("%s: socket create failed.\n", __func__);
        return -1;
    }

    memset(&sock_addr, 0, sizeof(struct sockaddr_un));
    sock_addr.sun_family = AF_UNIX;
    strncpy(sock_addr.sun_path,
            SOCK_PATH,
            sizeof(sock_addr.sun_path)-1);

    while(connect(m_nSocketfd,(const struct sockaddr*)&sock_addr,sizeof(struct sockaddr_un))) {
        if(wait_loop < 60) {
            if(!wait_loop)
                printf("Waiting for argus Server producer ");
            else
                printf(".");
            fflush(stdout);
            sleep(1);
            wait_loop++;
        } else {
            printf("\n%s: Waiting timed out\n", __func__);
			close(m_nSocketfd);
            return -1;
        }
    }
    if(wait_loop)
        printf("\n");
    return m_nSocketfd;
}


int CrossUtils::ancil_send_fds_with_buffer(int sock, const int *fds, unsigned n_fds, void *buffer)
{
    struct msghdr msghdr;
    struct iovec nothing_ptr;
    struct cmsghdr *cmsg;
    int i;
    
    nvbuf_param par;
    NvBufferGetParams(fds[0], &par.params);
    NvBufferGetParamsEx(fds[0], &par.paramsEx);
    //printf("ancil send fds with buffer,n_fds:%d,fds[0]:%d,w:%d,h:%d\n",n_fds,fds[0],par.params.width[0],par.params.height[0]);
    nothing_ptr.iov_base =&par;//&sendbuf;//&nothing;
    nothing_ptr.iov_len = sizeof(par);//sizeof(sendbuf);//1;
    msghdr.msg_name = NULL;
    msghdr.msg_namelen = 0;
    msghdr.msg_iov = &nothing_ptr;
    msghdr.msg_iovlen = 1;
    msghdr.msg_flags = 0;
    msghdr.msg_control = buffer;
    msghdr.msg_controllen = sizeof(struct cmsghdr) + sizeof(int) * n_fds;
    cmsg = CMSG_FIRSTHDR(&msghdr);
    cmsg->cmsg_len = msghdr.msg_controllen;
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
    for(i = 0; i < n_fds; i++)
	((int *)CMSG_DATA(cmsg))[i] = fds[i];
    return(sendmsg(sock, &msghdr, 0) >= 0 ? 0 : -1);	
}

int CrossUtils::Get_Flag(int index)
{
    return m_nFlag[index];
}


int CrossUtils::ancil_recv_fds_with_buffer(int sock, int *fds, unsigned n_fds, void *buffer)
{
    struct msghdr msghdr;
    struct iovec nothing_ptr;
    struct cmsghdr *cmsg;
    int i;
    struct timeval tv1, tv2;

    nvbuf_param par;
    nothing_ptr.iov_base =&par;//&recvbuf;//&nothing;
    nothing_ptr.iov_len =sizeof(par);//sizeof(recvbuf);//1;
    msghdr.msg_name = NULL;
    msghdr.msg_namelen = 0;
    msghdr.msg_iov = &nothing_ptr;
    msghdr.msg_iovlen = 1;
    msghdr.msg_flags = 0;
    //printf("###ancil recv fds with buffer[%d]\n",__LINE__);    
    msghdr.msg_control = buffer;
    msghdr.msg_controllen = sizeof(struct cmsghdr) + sizeof(int) * n_fds;
    cmsg = CMSG_FIRSTHDR(&msghdr);
    cmsg->cmsg_len = msghdr.msg_controllen;
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
    for(i = 0; i < n_fds; i++)
		((int *)CMSG_DATA(cmsg))[i] = -1;
    //printf("###ancil recv fds with buffer[%d],msg_controllen:%ld\n",__LINE__,msghdr.msg_controllen);    
    if(recvmsg(sock, &msghdr, 0) < 0)
    {
        printf("###ancil recv fds with buffer recvmsg error[%d]\n",__LINE__);    
        perror("recvmsg");    
	    return -1;
    }
    //printf("###ancil recv fds with buffer[%d],w:%d,h:%d\n",__LINE__,par.params.width[0],par.params.height[0]);    
    for(i = 0; i < n_fds; i++)
		fds[i] = ((int *)CMSG_DATA(cmsg))[i];
    n_fds = (msghdr.msg_controllen - sizeof(struct cmsghdr)) / sizeof(int);
    //printf("###ancil recv fds with buffer[%d],n_fds:%d,fds:%d\n",__LINE__,n_fds,fds[0]);    
    
    NvBufferParams dst_params;
    NvBufferParamsEx dst_paramsEx;

    int nDestFD = -1;
    NvBufferCreateParams input_params = {0};
    input_params.payloadType = NvBufferPayload_SurfArray;
    input_params.width = par.params.width[0];
    input_params.height = par.params.height[0];
    input_params.layout = NvBufferLayout_Pitch;
    input_params.nvbuf_tag = NvBufferTag_VIDEO_CONVERT;
    input_params.colorFormat = NvBufferColorFormat_YUV420;

    gettimeofday(&tv1, NULL);
    if (NvBufferCreateEx(&nDestFD, &input_params) < 0)
    {
        printf("CVideoFeeder::ConvertFD()->NvBufferCreateEx() fail!\n");
        return -1;
    }
    NvBufferGetParams(nDestFD, &dst_params);
    NvBufferGetParamsEx(nDestFD, &dst_paramsEx);

    NvBufferTransformParams transform_params={0};
    transform_params.transform_flag = NVBUFFER_TRANSFORM_FILTER;
    transform_params.transform_flip = NvBufferTransform_None; 
    transform_params.transform_filter = NvBufferTransform_Filter_Smart;
    transform_params.src_rect.top = 0;
    transform_params.src_rect.left = 0;
    transform_params.src_rect.width = par.params.width[0];
    transform_params.src_rect.height = par.params.height[0];
    transform_params.dst_rect.top = 0;
    transform_params.dst_rect.left = 0;
    transform_params.dst_rect.width = par.params.width[0];
    transform_params.dst_rect.height = par.params.height[0];
    if(NvBufferTransformEx(fds[0], &par.paramsEx,nDestFD, &dst_paramsEx, &transform_params)==0)
    {
        gettimeofday(&tv2, NULL);
        m_nCount++;
        printf("nvbuffer transforex success,diff[%d],count[%d]\n",get_time_diff_ms(&tv1, &tv2),m_nCount);
        NvBufferDestroy(nDestFD);
        //NvBufferDestroy(fds[0]);
#if 0
        NvJPEGEncoder* pJpegEnc = NvJPEGEncoder::createJPEGEncoder("jpenenc");
        if (pJpegEnc == NULL)
        {
            printf("NvJPEGEncoder::createJPEGEncoder() fail!\n");
            return -1;
        }
        pJpegEnc->setCropRect(0, 0,par.params.width[0],par.params.height[0]);
        unsigned long out_buf_size=par.params.width[0]*par.params.height[0]*3/2;
        unsigned char *out_buf = new unsigned char[out_buf_size];
        if (out_buf == NULL)
        {
            printf("out_buf NULL failed\n");
            NvBufferDestroy(nDestFD);
            return -1;
        }
        if (pJpegEnc->encodeFromFd(nDestFD, JCS_YCbCr, &out_buf, out_buf_size, 75) < 0)
        {
            printf("pJpegEnc->encodeFromFd() fail!\n");
            delete[] out_buf;
            NvBufferDestroy(nDestFD);
            return -1;
        }
        NvBufferDestroy(nDestFD);
        m_nCount++;
        if(m_nCount==10)
        {
            FILE * fp = fopen("out.jpg", "w+");
            fwrite(out_buf,out_buf_size,1,fp);
            fclose(fp);
        }
        printf("###%s:recv count:%d\n",__func__,m_nCount);
        delete[] out_buf;
#endif
    }
    else
    {
        printf("nvbuffer transforex failed\n");
    }
    //if(recvmsg(sock, &msghdr, 0) < 0)
    sendmsg(sock,&msghdr, 0);
	return 0;
}

int CrossUtils::ancil_recv_fd(int sock,int *fd)
{
    ANCIL_FD_BUFFER(1) buffer;

    return(ancil_recv_fds_with_buffer(sock, fd, 1, &buffer) == 1 ? 0 : -1);	
}

int CrossUtils::ancil_send_fd(int fd)
{
    int i;
    bool flag=false;
    ANCIL_FD_BUFFER(1) buffer;
	while(1){
		if(m_nSocketfd<0){
			printf("%s:ServerCreate  fd not init\n",__func__);
			sleep(3);
			continue;
		}
		else{
			//for(i=0; i<MAX_CLIENT; i++){
				if(m_nClientfd[0]>0){
					//printf("%s:ServerCreate client connect\n",__func__);
                    break;
				}
				else{
					//printf("%s:ServerCreate client not connect\n",__func__);
				}					
			//}
		}
	}
    return(ancil_send_fds_with_buffer(m_nClientfd[0], &fd, 1, &buffer));	
}

int CrossUtils::send_fd(int fd,nvbuf_param *par)
{
    struct msghdr msghdr;
    struct iovec nothing_ptr;
    struct cmsghdr *cmsg;
	//char ctrl_buf[CMSG_SPACE(sizeof(int))];
	char ctrl_buf[1024];
	int i;
	
    nothing_ptr.iov_base =par;
    nothing_ptr.iov_len = sizeof(nvbuf_param);
    msghdr.msg_name = NULL;
    msghdr.msg_namelen = 0;
    msghdr.msg_iov = &nothing_ptr;
    msghdr.msg_iovlen = 1;
    msghdr.msg_flags = 0;
    msghdr.msg_control = ctrl_buf;
    msghdr.msg_controllen = sizeof(struct cmsghdr) + sizeof(int);
    cmsg = CMSG_FIRSTHDR(&msghdr);
    cmsg->cmsg_len = msghdr.msg_controllen;
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
	*(int *)CMSG_DATA(cmsg) = fd;
	for(i=0; i<MAX_CLIENT; i++){
		if(m_nClientfd[i]>0){
			sendmsg(m_nClientfd[i],&msghdr,0);	
		}
	}
    return 0;
}

int CrossUtils::recv_fd(int sock,nvbuf_param *par)
{  
    struct msghdr msghdr;
    struct iovec nothing_ptr;
    struct cmsghdr *cmsg;
	//char ctrl_buf[CMSG_SPACE(sizeof(int))];  
	char ctrl_buf[1024];  
	int  fd;
    //nvbuf_param par;
    nothing_ptr.iov_base =par;
    nothing_ptr.iov_len =sizeof(nvbuf_param);
    msghdr.msg_name = NULL;
    msghdr.msg_namelen = 0;
    msghdr.msg_iov = &nothing_ptr;
    msghdr.msg_iovlen = 1;
    msghdr.msg_flags = 0;    
    msghdr.msg_control = ctrl_buf;
    msghdr.msg_controllen = sizeof(struct cmsghdr) + sizeof(int);
    cmsg = CMSG_FIRSTHDR(&msghdr);
    cmsg->cmsg_len = msghdr.msg_controllen;
    cmsg->cmsg_level = SOL_SOCKET;
    cmsg->cmsg_type = SCM_RIGHTS;
	*(int *)CMSG_DATA(cmsg) = -1;
    printf("###ancil recv fds with buffer[%d],msg_controllen:%ld\n",__LINE__,msghdr.msg_controllen);    
    if(recvmsg(sock, &msghdr,MSG_DONTWAIT) < 0)
    {  
        perror("recvmsg");    
	    return -1;
    }
	fd= *(int *)CMSG_DATA(cmsg);
    return fd;
}


void* CrossUtils::ServerThreadFun(void* pParam)
{
	struct sockaddr_un client_address;
	socklen_t address_len;
	fd_set ser_fdset;
	int client_sock_fd;
	int max_fd=1;
	int i;
	int ret;
	int flags=-1;
	struct timeval mytime;
	if(pParam == NULL)
    {
        return NULL;
    }
	CrossUtils* pThis = (CrossUtils*)pParam;
	
    pThis->ServerCreate();
	printf("%s:wait for client connnect\n",__func__);
	while(1)
	{
		if(pThis->m_nSocketfd<0)
		{
			printf("%s:ServerCreate failed\n",__func__);
			sleep(3);
		    pThis->ServerCreate();
			continue;
		}
		mytime.tv_sec=0;
        mytime.tv_usec=500000;
        FD_ZERO(&ser_fdset);
		FD_SET(pThis->m_nSocketfd,&ser_fdset);
	    //printf("%s:fd_set:%d\n",__func__,pThis->m_nSocketfd);
		if(max_fd < pThis->m_nSocketfd)
        {
            max_fd = pThis->m_nSocketfd;
			printf("%s:max_fd:%d\n",__func__,max_fd);
        }
		for(i=0;i<MAX_CLIENT;i++)  
        {
            if(pThis->m_nClientfd[i]>0)
            {
                FD_SET(pThis->m_nClientfd[i],&ser_fdset);
			    //printf("%s:fd_set:%d-%d\n",__func__,pThis->m_nClientfd[i],i);
                if(max_fd < pThis->m_nClientfd[i])
                {
                    max_fd = pThis->m_nClientfd[i];
			        printf("%s:max_fd:%d\n",__func__,max_fd);
                }
            }
        }
		ret = select(max_fd+1,&ser_fdset, NULL, NULL,&mytime);
		//ret = select(pThis->m_nSocketfd+1,&ser_fdset, NULL, NULL,&mytime);
        if(ret < 0)   
        {   
            perror("select failure\n");   
            continue;   
        }
		else if(0==ret)
		{
			//printf("%s:select time out\n",__func__);
			continue;
		}
		else
		{
			//printf("%s:select ok,ret:%d\n",__func__,ret);
			if(FD_ISSET(pThis->m_nSocketfd, &ser_fdset))
            {
			    printf("%s:select FD_ISSET m_nSocketfd\n",__func__);
                client_sock_fd = accept(pThis->m_nSocketfd,(struct sockaddr *)&client_address, &address_len);
                if(client_sock_fd > 0)
                {
                    for(i=0;i<MAX_CLIENT;i++)
                    {
                        if(pThis->m_nClientfd[i] == 0)
                        {
                            flags=i;
                            pThis->m_nClientfd[i] = client_sock_fd;
                            break;
                        }
                    }
                    if (flags >= 0)
                    {
						printf("%s:new user client[%d-%d] add sucessfully\n",__func__,flags,client_sock_fd);
                    }
                    else 
                    {  
						printf("%s:the client is full\n",__func__);
                    }
                }   
            }
			for(i=0; i<MAX_CLIENT; i++)
			{
				if(pThis->m_nClientfd[i]>0)
				{
					if(FD_ISSET(pThis->m_nClientfd[i],&ser_fdset))
					{
						//printf("%s:client FD_ISSET[%d-%d]\n",__func__,i,pThis->m_nClientfd[i]);
						char buf[1024];
						ret=read(pThis->m_nClientfd[i],buf,sizeof(buf));
						//printf("%s:client read[%d-%d] ret:%d-%d\n",__func__,i,pThis->m_nClientfd[i],ret,errno);
                        if(ret==0)
                        {
                            FD_CLR(pThis->m_nClientfd[i], &ser_fdset);
                            pThis->m_nClientfd[i]=0;
                        }
                        else
                        {
                            printf("client[%d] ack[%d-%d] \t",i,ret,pThis->m_nFlag[i]++);
                        }

					}
				}
			} 
		}	
	}
	printf("%s:exit\n",__func__);
	return NULL;
}

